/* The el_* functions in a version that doesn't require as humongous
   a TextColorTable, while still being (hopefully) fastish, for PalmOS.
   
   Originally, TextColorTable contained 8 8-bit pixels for each combination
   of forecolor/backcolor/data (total of 512 kB).
   
   This version contains 8 4-bit pixels for foreground color ONLY.
   To get both forecolor and backcolor, do another lookup with ~data and the
   backcolor, then | the two together.
   This way, we need only 16 kB, which is acceptable.

   Whoops, scrolling adds a problem when the CPU is wrong-endian.
   
   The (4-bit) pixels 12345678
   reside in a uint32 like this:
   0x78563412
   When scrolling an even number of pixels (whole bytes) that's not a problem:
       pl          pl2       x_scroll=4
   0x78563412  0x78563412
   0x7856     |    0x34120000 =
        0x34127856           
            out

   Scrolling an odd number of pixels IS a problem:
       pl          pl2       x_scroll=3
   0x78563412  0x78563412
   0x785      |   0x634120000 =
        0x63412785        
            out WRONG! correct is 0x45238167

   I added table elements that have been prescrolled one pixel.
       pl          pl2       x_scroll=3
   0x67452381  0x67452381
   0x67       |  0x4523010000
      |  0x8000 =
        0x45238167

   Of course it turned out that most 16 color graphics hardware have a different nybble order
   So I added yet another set of table elements for use when pixels-in-a-int look like this
   0x87654321.
   
   And yet another set of elements for 8bit mode
   0x04030201.

   Now TextColorTable is 49 kB. Still, that's better than 512.
   
   Assumptions in the below code: COL80_XSTART must be divisible by 8,
   p points at col80 (scrolling is handled inside the el_* functions).
*/

void MOS6569::el_std_text(uint8 *p, uint8 *q, uint8 *r)
{
	switch (the_c64->armletData->prefs->ScreenMode) {
		case SCREENMODE_16:
		case SCREENMODE_16ALT:
			el_std_text_16(p, q, r,
				the_c64->armletData->prefs->ScreenMode==SCREENMODE_16ALT);
			break;
		default:
			el_std_text_256(p, q, r);
	}
}
void MOS6569::el_std_text_16(uint8 *p, uint8 *q, uint8 *r, uint8 alt)
{
        uint32 b0cc = b0c & 0xf;
	uint32 *lp = (uint32 *)p;
	uint8 sh = (x_scroll&(alt?6:7)) << 2, sh2 = 32-sh;
	uint8 odd = x_scroll&1;
	uint8 tctpart = alt? odd:2;
	uint32 pl = b0c_color_long, pl2;
	uint8 *cp = color_line;
	uint8 *mp = matrix_line;
	
	// Loop for 40 characters
	//ShowRequester(REQ_EL_TEXT, (int32) p, (int32) mp, (int32) sh, 0);
	for (int i=0; i<40; i++) {
		uint8 color = cp[i] & 0xf;
		uint8 data = r[i] = q[mp[i] << 3];

		pl2 = TextColorTable[(tctpart<<12) | (color<<8) | data] |
                      TextColorTable[(tctpart<<12) | (b0cc<<8)  | ((~data)&0xff)];
		if (odd && alt)
			*lp++ = (pl>>sh2) | (((pl2 & 0xffffff0f) | (pl & 0xf0))<<sh);
		else
			*lp++ = (pl>>sh2) | (pl2<<sh);
		pl = pl2;
	}
}
void MOS6569::el_std_text_256(uint8 *p, uint8 *q, uint8 *r)
{
        uint32 b0cc = b0c & 0xf;
	uint32 *lp = (uint32 *)p;
	uint8 sh = (x_scroll&7) << 3, sh2 = 64-sh;
	uint8 tctpart = 3;
	uint32 pla, plb, pl2a, pl2b;
	uint8 *cp = color_line;
	uint8 *mp = matrix_line;

	pla=plb=b0c | (b0c<<8) | (b0c<<16) | (b0c<<24);

	// Loop for 40 characters
	//ShowRequester(REQ_EL_TEXT, (int32) p, (int32) mp, (int32) sh, 0);
	for (int i=0; i<40; i++) {
		uint8 color = cp[i] & 0xf;
		uint8 data = r[i] = q[mp[i] << 3];

		// First 4 pixels
		pl2a = TextColorTable[(tctpart<<12) | (color<<4) | (data>>4)] |
                       TextColorTable[(tctpart<<12) | (b0cc<<4)  | ((~(data>>4))&0xf)];
		// Other 4 pixles
		pl2b = TextColorTable[(tctpart<<12) | (color<<4) | (data&0xf)] |
                       TextColorTable[(tctpart<<12) | (b0cc<<4)  | ((~data)&0xf)];

		//    pla      plb      pl2a    pl2b
		// 04030201 08070605 12111009 16151413
		if (sh&32)  {
			// sh=56, sh2=8
			// 040302 | 05000000 = 05040302
			*lp++ = (pla>>sh2) | (plb<<(sh-32));
			// 080706 | 09000000 = 09080706
			*lp++ = (plb>>sh2) | (pl2a<<(sh-32));
			// in memory 0203040506070809
		} else {
			// sh=8, sh2=56
			// 08 | 11100900 = 11100908
			*lp++ = (plb>>(sh2-32)) | (pl2a<<sh);
			// 12 | 15141300 = 15141312
			*lp++ = (pl2a>>(sh2-32)) | (pl2b<<sh);
			// in memory 0809101112131415
		}
		pla = pl2a;
		plb = pl2b;
	}
}


void MOS6569::el_mc_text(uint8 *p, uint8 *q, uint8 *r)
{
	switch (the_c64->armletData->prefs->ScreenMode) {
		case SCREENMODE_16:
		case SCREENMODE_16ALT:
			el_mc_text_16(p, q, r,
				the_c64->armletData->prefs->ScreenMode==SCREENMODE_16ALT);
			break;
		default:
			el_mc_text_256(p, q, r);
	}
}
void MOS6569::el_mc_text_16(uint8 *p, uint8 *q, uint8 *r, uint8 alt)
{
	uint32 *lp = (uint32 *)p;
	uint8 *cp = color_line;
	uint8 *mp = matrix_line;
	uint16 *mclp = mc_color_lookup;
	uint8 sh = (x_scroll&(alt?6:7)) << 2, sh2 = 32-sh, odd=x_scroll&1;
	uint32 pl = b0c_color_long, pl2;
	uint8 tctpart=alt? odd:2;

	//ShowRequester(REQ_EL_TEXT, (int32) p, (int32) mp, (int32) sh, 0); 
	// Loop for 40 characters
	for (int i=0; i<40; i++) {
		uint8 data = q[mp[i] << 3];

		if (cp[i] & 8) {
			uint8 color = cp[i]&7;
			r[i] = (data & 0xaa) | (data & 0xaa) >> 1;
			mclp[3] = color | (color << 4);
			pl2 = (mclp[(data >> 6) & 3]      ) |
			      (mclp[(data >> 4) & 3] <<  8) |
			      (mclp[(data >> 2) & 3] << 16) |
			      (mclp[ data       & 3] << 24);
			
			// pl=44332211  pl2=88776655
			// scroll=3, sh=8, sh2=24
			if (alt && odd) {
				// OMAP lcd
				// Desired result in memory 34455667
				// in uint32 67564534
				//     00004433&0xf0f0f0f0 = 00004030
				*lp++ = ((pl>>(sh2-8))&0xf0f0f0f0) |
				//     00000044&0x0f0f0f0f = 00000004
				      ((pl>>sh2)&0x0f0f0f0f) |
				//     66550000&0xf0f0f0f0 = 60500000
				      ((pl2<<(sh+8))&0xf0f0f0f0) |
				//     77665500&0x0f0f0f0f = 07060500
				      ((pl2<<sh)&0x0f0f0f0f);
				//                           67564534
			} else {
				// XScale lcd, reverse nybble order
				// Desired result in memory 43546576
				// in uint32 76655443
				*lp++ = (pl>>sh2) | (pl2<<sh);
			}
		} else { // Standard mode in multicolor mode
			uint8 color = cp[i];
			r[i] = data;

			pl2 = TextColorTable[(tctpart<<12) | (color<<8) | data] |
                              TextColorTable[(tctpart<<12) | (b0c<<8) | ((~data)&0xff)];
			if (odd && alt)
				// The table saves us some pain
				*lp++ = (pl>>sh2) | (((pl2 & 0xffffff0f) | (pl & 0xf0))<<sh);
			else
				*lp++ = (pl>>sh2) | (pl2<<sh);
		}
		
		pl = pl2;
	}
}
void MOS6569::el_mc_text_256(uint8 *p, uint8 *q, uint8 *r)
{
	uint32 *lp = (uint32 *)p;
	uint8 *cp = color_line;
	uint8 *mp = matrix_line;
	uint16 mclp[4];
	uint8 sh = (x_scroll&7) << 3, sh2 = 64-sh;
	uint32 pla, plb, pl2a, pl2b;
	uint8 tctpart=3, i;

	pla=plb=b0c | (b0c<<8) | (b0c<<16) | (b0c<<24);

	for (i=0; i<3; i++)
		mclp[i] = ((mc_color_lookup[i]<<4)&0xf00) |
		           (mc_color_lookup[i]&0x0f);

	// Loop for 40 characters
	for (int i=0; i<40; i++) {
		uint8 data = q[mp[i] << 3];

		if (cp[i] & 8) {
			uint8 color = cp[i]&7;
			r[i] = (data & 0xaa) | (data & 0xaa) >> 1;
			mclp[3] = color | (color << 8);
			pl2a = (mclp[(data >> 6) & 3]      ) |
			       (mclp[(data >> 4) & 3] << 16);
			pl2b = (mclp[(data >> 2) & 3]      ) |
			       (mclp[ data       & 3] << 16);

		} else { // Standard mode in multicolor mode
			uint8 color = cp[i];
			r[i] = data;

			pl2a = TextColorTable[(tctpart<<12) | (color<<4) | ((data>>4)&0xf)] |
                               TextColorTable[(tctpart<<12) | (b0c<<4) | ((~(data>>4))&0xf)];
			pl2b = TextColorTable[(tctpart<<12) | (color<<4) | (data&0xf)] |
                               TextColorTable[(tctpart<<12) | (b0c<<4) | ((~data)&0xf)];
		}
		
		if (sh&32)  {
			*lp++ = (pla>>sh2) | (plb<<(sh-32));
			*lp++ = (plb>>sh2) | (pl2a<<(sh-32));
		} else {
			*lp++ = (plb>>(sh2-32)) | (pl2a<<sh);
			*lp++ = (pl2a>>(sh2-32)) | (pl2b<<sh);
		}
		pla = pl2a;
		plb = pl2b;
	}
}

void MOS6569::el_std_bitmap(uint8 *p, uint8 *q, uint8 *r)
{
	switch (the_c64->armletData->prefs->ScreenMode) {
		case SCREENMODE_16:
		case SCREENMODE_16ALT:
			el_std_bitmap_16(p, q, r,
				the_c64->armletData->prefs->ScreenMode==SCREENMODE_16ALT);
			break;
		default:
			el_std_bitmap_256(p, q, r);
	}
}
void MOS6569::el_std_bitmap_16(uint8 *p, uint8 *q, uint8 *r, uint8 alt)
{
	uint32 *lp = (uint32 *)p;
	uint8 *mp = matrix_line;
	uint8 sh = (x_scroll&(alt?6:7)) << 2, sh2 = 32-sh, odd=x_scroll&1;
	uint8 tctpart=alt? odd:2;
	uint32 pl = b0c_color_long, pl2;

	//ShowRequester(REQ_EL_GFX, (int32) p, (int32) mp, (int32) sh, 0); 
	// Loop for 40 characters
	for (int i=0; i<40; i++, q+=8) {
		uint8 data = r[i] = *q;
		uint8 color = mp[i] >> 4;
		uint8 bcolor = mp[i] & 15;

		pl2 = TextColorTable[(tctpart<<12) | (color<<8) | data] |
                      TextColorTable[(tctpart<<12) | (bcolor<<8) | ((~data)&0xff)];
		if (odd && alt)
			*lp++ = (pl>>sh2) | (((pl2 & 0xffffff0f) | (pl & 0xf0))<<sh);
		else
			*lp++ = (pl>>sh2) | (pl2<<sh);

		pl = pl2;
	}
}
void MOS6569::el_std_bitmap_256(uint8 *p, uint8 *q, uint8 *r)
{
	uint32 *lp = (uint32 *)p;
	uint8 *mp = matrix_line;
	uint8 sh = (x_scroll&7) << 3, sh2 = 64-sh;
	uint32 pla, plb, pl2a, pl2b;
	uint8 tctpart=3;

	pla=plb=b0c | (b0c<<8) | (b0c<<16) | (b0c<<24);

	//ShowRequester(REQ_EL_GFX, (int32) p, (int32) mp, (int32) sh, 0); 
	// Loop for 40 characters
	for (int i=0; i<40; i++, q+=8) {
		uint8 data = r[i] = *q;
		uint8 color = mp[i] >> 4;
		uint8 bcolor = mp[i] & 15;

		pl2a = TextColorTable[(tctpart<<12) | (color<<4) | ((data>>4)&0xf)] |
                       TextColorTable[(tctpart<<12) | (bcolor<<4) | ((~(data>>4))&0xf)];
		pl2b = TextColorTable[(tctpart<<12) | (color<<4) | (data&0xf)] |
                       TextColorTable[(tctpart<<12) | (bcolor<<4) | ((~data)&0xf)];

		if (sh&32)  {
			*lp++ = (pla>>sh2) | (plb<<(sh-32));
			*lp++ = (plb>>sh2) | (pl2a<<(sh-32));
		} else {
			*lp++ = (plb>>(sh2-32)) | (pl2a<<sh);
			*lp++ = (pl2a>>(sh2-32)) | (pl2b<<sh);
		}
		pla = pl2a;
		plb = pl2b;
	}
}


void MOS6569::el_mc_bitmap(uint8 *p, uint8 *q, uint8 *r)
{
	switch (the_c64->armletData->prefs->ScreenMode) {
		case SCREENMODE_16:
		case SCREENMODE_16ALT:
			el_mc_bitmap_16(p, q, r,
				the_c64->armletData->prefs->ScreenMode==SCREENMODE_16ALT);
			break;
		default:
			el_mc_bitmap_256(p, q, r);
	}
}

void MOS6569::el_mc_bitmap_16(uint8 *p, uint8 *q, uint8 *r, uint8 alt)
{
	uint8 lookup[4];
	uint32 *lp = (uint32 *) p;
	uint8 *cp = color_line;
	uint8 *mp = matrix_line;
	uint8 sh = (x_scroll&(alt?6:7)) << 2, sh2 = 32-sh, odd=x_scroll&1;
	uint32 pl = b0c_color_long, pl2;

#ifdef __GNU_C__
	&lookup; /* Statement with no effect other than preventing GCC from
			  * putting the array in a register, which generates
			  * spectacularly bad code. */
#endif

	//ShowRequester(REQ_EL_GFX, (int32) p, (int32) mp, (int32) sh, 0); 

	lookup[0] = (b0c_color << 4) | b0c_color;

	// Loop for 40 characters
	for (int i=0; i<40; i++, q+=8) {
		uint8 color, acolor, bcolor;

		color = mp[i] >> 4;
		lookup[1] = (color << 4) | color;
		bcolor = mp[i] & 0xf;
		lookup[2] = (bcolor << 4) | bcolor;
		acolor = cp[i] & 0xf;
		lookup[3] = (acolor << 4) | acolor;

		uint8 data = *q;
		r[i] = (data & 0xaa) | (data & 0xaa) >> 1;

		pl2 = (lookup[(data >> 6) & 3]      ) |
		      (lookup[(data >> 4) & 3] <<  8) |
		      (lookup[(data >> 2) & 3] << 16) |
		      (lookup[ data       & 3] << 24);
			
		if (alt && odd) {
			*lp++ = ((pl>>(sh2-8))&0xf0f0f0f0) |
			        ((pl>>sh2)&0x0f0f0f0f) |
			        ((pl2<<(sh+8))&0xf0f0f0f0) |
			        ((pl2<<sh)&0x0f0f0f0f);
		} else {
			*lp++ = (pl>>sh2) | (pl2<<sh);
		}
		pl=pl2;
	}
}

void MOS6569::el_mc_bitmap_256(uint8 *p, uint8 *q, uint8 *r)
{
	uint16 lookup[4];
	uint32 *lp = (uint32 *) p;
	uint8 *cp = color_line;
	uint8 *mp = matrix_line;
	uint8 sh = (x_scroll&7) << 3, sh2 = 64-sh;
	uint8 tctpart=3;
	uint32 pla, plb, pl2a, pl2b;

	pla=plb=(b0c_color<<24)|(b0c_color<<16)|(b0c_color<<8)|b0c_color; 

#ifdef __GNU_C__
	&lookup; /* Statement with no effect other than preventing GCC from
			  * putting the array in a register, which generates
			  * spectacularly bad code. */
#endif

	//ShowRequester(REQ_EL_GFX, (int32) p, (int32) mp, (int32) sh, 0); 

	lookup[0] = (b0c_color << 8) | b0c_color;

	// Loop for 40 characters
	for (int i=0; i<40; i++, q+=8) {
		uint8 color, acolor, bcolor;

		color = mp[i] >> 4;
		lookup[1] = (color << 8) | color;
		bcolor = mp[i] & 0xf;
		lookup[2] = (bcolor << 8) | bcolor;
		acolor = cp[i] & 0xf;
		lookup[3] = (acolor << 8) | acolor;

		uint8 data = *q;
		r[i] = (data & 0xaa) | (data & 0xaa) >> 1;

		pl2a = (lookup[(data >> 6) & 3]      ) |
		       (lookup[(data >> 4) & 3] << 16);
		pl2b = (lookup[(data >> 2) & 3]      ) |
		       (lookup[ data       & 3] << 16);
			
		if (sh&32)  {
			*lp++ = (pla>>sh2) | (plb<<(sh-32));
			*lp++ = (plb>>sh2) | (pl2a<<(sh-32));
		} else {
			*lp++ = (plb>>(sh2-32)) | (pl2a<<sh);
			*lp++ = (pl2a>>(sh2-32)) | (pl2b<<sh);
		}
		pla = pl2a;
		plb = pl2b;
	}
}


void MOS6569::el_ecm_text(uint8 *p, uint8 *q, uint8 *r)
{
	switch (the_c64->armletData->prefs->ScreenMode) {
		case SCREENMODE_16:
		case SCREENMODE_16ALT:
			el_ecm_text_16(p, q, r,
				the_c64->armletData->prefs->ScreenMode==SCREENMODE_16ALT);
			break;
		default:
			el_ecm_text_256(p, q, r);
	}
}
void MOS6569::el_ecm_text_16(uint8 *p, uint8 *q, uint8 *r, uint8 alt)
{
	uint32 *lp = (uint32 *)p;
	uint8 *cp = color_line;
	uint8 *mp = matrix_line;
	uint8 *bcp = &b0c;
	uint8 sh = (x_scroll&(alt?6:7)) << 2, sh2 = 32-sh, odd=x_scroll&1;
	uint8 tctpart = alt? odd:2;
	uint32 pl = b0c_color_long, pl2;

	//ShowRequester(REQ_EL_TEXT, (int32) p, (int32) mp, (int32) sh, 0); 
	// Loop for 40 characters
	for (int i=0; i<40; i++) {
		uint8 data = r[i] = mp[i];
		uint8 color = cp[i];
		uint8 bcolor = bcp[(data >> 6) & 3];

		data = q[(data & 0x3f) << 3];
		pl2 = TextColorTable[(tctpart<<12) | (color<<8) | data] |
                      TextColorTable[(tctpart<<12) | (bcolor<<8) | ((~data)&0xff)];
		if (odd && alt)
			*lp++ = (pl>>sh2) | (((pl2 & 0xffffff0f) | (pl & 0xf0))<<sh);
		else
			*lp++ = (pl>>sh2) | (pl2<<sh);
		pl = pl2;
	}
}
void MOS6569::el_ecm_text_256(uint8 *p, uint8 *q, uint8 *r)
{
	uint32 *lp = (uint32 *)p;
	uint8 *cp = color_line;
	uint8 *mp = matrix_line;
	uint8 *bcp = &b0c;
	uint8 sh = (x_scroll&7) << 3, sh2 = 64-sh;
	uint8 tctpart = 3;
	uint32 pla, plb, pl2a, pl2b;

	pla=plb=(b0c<<24)|(b0c<<16)|(b0c<<8)|b0c;

	//ShowRequester(REQ_EL_TEXT, (int32) p, (int32) mp, (int32) sh, 0); 
	// Loop for 40 characters
	for (int i=0; i<40; i++) {
		uint8 data = r[i] = mp[i];
		uint8 color = cp[i];
		uint8 bcolor = bcp[(data >> 6) & 3];

		data = q[(data & 0x3f) << 3];
		pl2a = TextColorTable[(tctpart<<12) | (color<<4) | ((data>>4)&0xf)] |
                      TextColorTable[(tctpart<<12) | (bcolor<<4) | ((~(data>>4))&0xf)];
		pl2b = TextColorTable[(tctpart<<12) | (color<<4) | (data&0xf)] |
                       TextColorTable[(tctpart<<12) | (bcolor<<4) | ((~data)&0xf)];
			
		if (sh&32)  {
			*lp++ = (pla>>sh2) | (plb<<(sh-32));
			*lp++ = (plb>>sh2) | (pl2a<<(sh-32));
		} else {
			*lp++ = (plb>>(sh2-32)) | (pl2a<<sh);
			*lp++ = (pl2a>>(sh2-32)) | (pl2b<<sh);
		}
		pla = pl2a;
		plb = pl2b;
	}
}

void MOS6569::el_std_idle(uint8 *p, uint8 *r)
{
	switch (the_c64->armletData->prefs->ScreenMode) {
		case SCREENMODE_16:
		case SCREENMODE_16ALT:
			el_std_idle_16(p, r,
				the_c64->armletData->prefs->ScreenMode==SCREENMODE_16ALT);
			break;
		default:
			el_std_idle_256(p, r);
	}
}
void MOS6569::el_std_idle_16(uint8 *p, uint8 *r, uint8 alt)
{
	uint8 data = *get_physical(ctrl1 & 0x40 ? 0x39ff : 0x3fff);
	uint32 *lp = (uint32 *)p;
	uint8 sh = (x_scroll&(alt?6:7))<<2, sh2 = 32-sh, odd=x_scroll&1;
	uint8 tctpart = alt? odd:2;
	uint32 pl = b0c_color_long;
	uint32 pl2=TextColorTable[(tctpart<<12) | (b0c<<8) | ((~data)&0xff)];
	
	if (odd && alt)
		*lp++ = (pl>>sh2) | (((pl2 & 0xffffff0f) | (pl & 0xf0))<<sh);
	else
		*lp++ = (pl>>sh2) | (pl2<<sh);

	if (odd && alt)
		pl = (pl2>>sh2) | (((pl2 & 0xffffff0f) | (pl2 & 0xf0))<<sh);
	else
		pl = (pl2>>sh2) | (pl2<<sh);

	for (int i=1; i<40; i++) {
		*lp++ = pl;
		*r++ = data;
	}
}
void MOS6569::el_std_idle_256(uint8 *p, uint8 *r)
{
	uint8 data = *get_physical(ctrl1 & 0x40 ? 0x39ff : 0x3fff);
	uint32 *lp = (uint32 *)p;
	uint8 sh = (x_scroll&7)<<3, sh2 = 64-sh;
	uint8 tctpart = 3;
	uint32 pla, plb, pl2a, pl2b;

	pla=plb=(b0c<<24)|(b0c<<16)|(b0c<<8)|b0c;
	// TODO: chack if I messed up this mode. Yes I did.
	pl2a=TextColorTable[(tctpart<<12) | (b0c<<4) | ((~(data>>4))&0xf)];
	pl2b=TextColorTable[(tctpart<<12) | (b0c<<4) | ((~data)&0xf)];
	
	if (sh&32)  {
		*lp++ = (pla>>sh2) | (plb<<(sh-32));
		*lp++ = (plb>>sh2) | (pl2a<<(sh-32));
	} else {
		*lp++ = (plb>>(sh2-32)) | (pl2a<<sh);
		*lp++ = (pl2a>>(sh2-32)) | (pl2b<<sh);
	}
	if (sh&32)  {
		pla = (pl2a>>sh2) | (pl2b<<(sh-32));
		plb = (pl2b>>sh2) | (pl2a<<(sh-32));
	} else {
		plb = (pl2b>>(sh2-32)) | (pl2a<<sh);
		pla = (pl2a>>(sh2-32)) | (pl2b<<sh);
	}

	for (int i=1; i<40; i++) {
		*lp++ = pla;
		*lp++ = plb;
		*r++ = data;
	}
}

void MOS6569::el_mc_idle(uint8 *p, uint8 *r)
{
	switch (the_c64->armletData->prefs->ScreenMode) {
		case SCREENMODE_16:
		case SCREENMODE_16ALT:
			el_mc_idle_16(p, r,
				the_c64->armletData->prefs->ScreenMode==SCREENMODE_16ALT);
			break;
		default:
			el_mc_idle_256(p, r);
	}
}

void MOS6569::el_mc_idle_16(uint8 *p, uint8 *r, uint8 alt)
{
	uint16 data = *get_physical(0x3fff);
	uint32 *lp = (uint32 *)p;
	uint8 sh = (x_scroll&(alt?6:7)) << 2, sh2 = 32-sh, odd=x_scroll&1;

	uint8 lookup[4];
	lookup[0] = (b0c_color << 4) | b0c_color;
	lookup[1] = lookup[2] = lookup[3] = 0;

        uint32 conv =  lookup[(data >> 6) & 3]        |
                      (lookup[(data >> 4) & 3] <<  8) |
                      (lookup[(data >> 2) & 3] << 16) |
                      (lookup[data & 3]        << 24);

	uint32 pl;

	if (alt && odd) {
		*lp++ = ((b0c_color_long>>(sh2-8))&0xf0f0f0f0) |
		        ((b0c_color_long>>sh2)&0x0f0f0f0f) |
		        ((conv<<(sh+8))&0xf0f0f0f0) |
		        ((conv<<sh)&0x0f0f0f0f);
		pl = ((conv>>(sh2-8))&0xf0f0f0f0) |
		     ((conv>>sh2)&0x0f0f0f0f) |
		     ((conv<<(sh+8))&0xf0f0f0f0) |
		     ((conv<<sh)&0x0f0f0f0f);
	} else {
		*lp++ = (b0c_color_long>>sh2) | (conv<<sh);
		pl = (conv>>sh2) | (conv<<sh);
	}
	for (int i=0; i<40; i++) {
		*lp++ = pl;
		*r++ = data;
	}
}
void MOS6569::el_mc_idle_256(uint8 *p, uint8 *r)
{
	uint16 data = *get_physical(0x3fff);
	uint32 *lp = (uint32 *)p;
	uint8 sh = (x_scroll&7) << 3, sh2 = 64-sh;

	uint16 lookup[4];
	lookup[0] = (b0c_color << 8) | b0c_color;
	lookup[1] = lookup[2] = lookup[3] = 0;

        uint32 conva =  lookup[(data >> 6) & 3] |
                       (lookup[(data >> 4) & 3] << 16);
	uint32 convb =  lookup[(data >> 2) & 3] |
                       (lookup[data & 3]        << 16);

	uint32 pla, plb;

	if (sh&32) {
		*lp++ = (b0c_color_long>>sh2) | (b0c_color_long<<(sh-32));
		*lp++ = (b0c_color_long>>sh2) | (conva<<(sh-32));
		pla = (conva>>sh2) | (convb<<(sh-32));
		plb = (convb>>sh2) | (conva<<(sh-32));
	} else {
		*lp++ = (b0c_color_long>>sh2) | (conva<<(sh-32));
		*lp++ = (conva>>sh2) | (convb<<(sh-32));
		pla = (convb>>sh2) | (conva<<(sh-32));
		plb = (conva>>sh2) | (convb<<(sh-32));
	}
	for (int i=0; i<40; i++) {
		*lp++ = pla;
		*lp++ = plb;
		*r++ = data;
	}
}

// Sprites

#define PUT_PIXEL() \
if (q[i]) \
	spr_coll |= q[i] | sbit;\
else { \
	if (byte) {\
		p[i]=col;\
	} else if (odd && alt) {\
		if (i&1) \
			p[(i>>1)+1] = (p[(i>>1)+1]&0x0f) | (col<<4);\
		else\
			p[i>>1] = (p[i>>1]&0xf0) | col;\
	} else {\
		if (i&1) \
			p[i>>1] = (p[i>>1]&0xf0) | col;\
		else\
			p[i>>1] = (p[i>>1]&0x0f) | (col<<4);\
	}\
	q[i] = sbit;\
}


#ifdef GLOBAL_VARS
static void el_sprites(uint8 *chunky_ptr)
#else
void MOS6569::el_sprites(uint8 *chunky_ptr)
#endif
{
	int i;
	int snum, sbit;		// Sprite number/bit mask
	int spr_coll=0, gfx_coll=0;
	int xoffs = (C64DISPLAY_X-DISPLAY_X)/2; // 32 if the screen is 320 wide
	uint8 alt, byte;

	alt=the_c64->armletData->prefs->ScreenMode==SCREENMODE_16ALT;
	byte=is256(the_c64->armletData->prefs->ScreenMode);
	
	// Draw each active sprite
	for (snum=0, sbit=1; snum<8; snum++, sbit<<=1)
		if ((sprite_on & sbit) && mx[snum] < C64DISPLAY_X-32) {
			int spr_mask_pos;	// Sprite bit position in fore_mask_buf
			uint32 sdata, fore_mask;

			uint8 *p = chunky_ptr + (mx[snum]-xoffs)/2 + 4;
			uint8 *q = spr_coll_buf + mx[snum] + 8;
			if (byte)
				p=chunky_ptr + (mx[snum]-xoffs) + 8;

			uint8 *sdatap = get_physical(matrix_base[0x3f8 + snum] << 6 | mc[snum]);
                        sdata = (*sdatap << 24) | (*(sdatap+1) << 16) | (*(sdatap+2) << 8);
			uint8 color = spr_color[snum], col;

			spr_mask_pos = mx[snum] + 8 - x_scroll;
			
			uint8 *fmbp = fore_mask_buf + (spr_mask_pos / 8);
			int sshift = spr_mask_pos & 7;
                        fore_mask = (((*(fmbp+0) << 24) | (*(fmbp+1) << 16) | (*(fmbp+2) << 8)
				      | (*(fmbp+3))) << sshift) | (*(fmbp+4) >> (8-sshift));
			// Don't draw outside the screen buffer!
			uint8 xstart=0, xstop=24, xstop1, odd;
			if (mxe & sbit) {
				xstop=48;
				xstop1=32;
			}
			// DEBUG
			
			if (mx[snum]+8 < xoffs) xstart=xoffs-mx[snum]-8;
			if (mx[snum]+8-xoffs+xstop > DISPLAY_X) {
				xstop = DISPLAY_X-mx[snum]-8+xoffs;
				if (xstop<xstop1) xstop1=xstop;
			}
			odd=mx[snum]&1;

			if (mxe & sbit) {		// X-expanded
				if (mx[snum] >= C64DISPLAY_X-56)
					continue;

				uint32 sdata_l = 0, sdata_r = 0, fore_mask_r;
				fore_mask_r = (((*(fmbp+4) << 24) | (*(fmbp+5) << 16) | (*(fmbp+6) << 8)
						| (*(fmbp+7))) << sshift) | (*(fmbp+8) >> (8-sshift));
				if (mmc & sbit) {	// Multicolor mode
					uint32 plane0_l, plane0_r, plane1_l, plane1_r;

					// Expand sprite data
					sdata_l = MultiExpTable[sdata >> 24 & 0xff] << 16 | MultiExpTable[sdata >> 16 & 0xff];
					sdata_r = MultiExpTable[sdata >> 8 & 0xff] << 16;

					// Convert sprite chunky pixels to bitplanes
					plane0_l = (sdata_l & 0x55555555) | (sdata_l & 0x55555555) << 1;
					plane1_l = (sdata_l & 0xaaaaaaaa) | (sdata_l & 0xaaaaaaaa) >> 1;
					plane0_r = (sdata_r & 0x55555555) | (sdata_r & 0x55555555) << 1;
					plane1_r = (sdata_r & 0xaaaaaaaa) | (sdata_r & 0xaaaaaaaa) >> 1;

					// Collision with graphics?
					if ((fore_mask & (plane0_l | plane1_l)) || (fore_mask_r & (plane0_r | plane1_r))) {
						gfx_coll |= sbit;
						if (mdp & sbit)	{
							plane0_l &= ~fore_mask;	// Mask sprite if in background
							plane1_l &= ~fore_mask;
							plane0_r &= ~fore_mask_r;
							plane1_r &= ~fore_mask_r;
						}
					}

					// Paint sprite
					plane0_l<<=xstart; plane1_l<<=xstart;
					for (i=xstart; i<xstop1; i++, plane0_l<<=1, plane1_l<<=1) {
						if (plane1_l & 0x80000000) {
							if (plane0_l & 0x80000000)
								col = mm1_color;
							else
								col = color;
						} else {
							if (plane0_l & 0x80000000)
								col = mm0_color;
							else
								continue;
						}
						PUT_PIXEL();
					}
					if (xstart>32) {
						plane0_r<<=xstart-32;
						plane1_r<<=xstart-32;
					}
					for (; i<xstop; i++, plane0_r<<=1, plane1_r<<=1) {
						if (plane1_r & 0x80000000) {
							if (plane0_r & 0x80000000)
								col = mm1_color;
							else
								col = color;
						} else {
							if (plane0_r & 0x80000000)
								col = mm0_color;
							else
								continue;
						}
						PUT_PIXEL();
					}

				} else {			// Standard mode

					// Expand sprite data
					sdata_l = ((uint32)ExpTable[sdata >> 24 & 0xff]) << 16 | ExpTable[sdata >> 16 & 0xff];
					sdata_r = ((uint32)ExpTable[sdata >> 8 & 0xff]) << 16;
					// Collision with graphics?
					if ((fore_mask & sdata_l) || (fore_mask_r & sdata_r)) {
						gfx_coll |= sbit;
						if (mdp & sbit)	{
							sdata_l &= ~fore_mask;	// Mask sprite if in background
							sdata_r &= ~fore_mask_r;
						}
					}

					// Paint sprite
					col = color;
					sdata_l<<=xstart;
					for (i=xstart; i<xstop1; i++, sdata_l<<=1)
						if (sdata_l & 0x80000000) {
							PUT_PIXEL();
						}
					if (xstart>32)
						sdata_r<<=xstart-32;
					for (; i<xstop; i++, sdata_r<<=1)
						if (sdata_r & 0x80000000) {
							PUT_PIXEL();
						}
				}

			} else					// Unexpanded

				if (mmc & sbit) {	// Multicolor mode
					uint32 plane0, plane1;

					// Convert sprite chunky pixels to bitplanes
					plane0 = (sdata & 0x55555555) | (sdata & 0x55555555) << 1;
					plane1 = (sdata & 0xaaaaaaaa) | (sdata & 0xaaaaaaaa) >> 1;

					// Collision with graphics?
					if (fore_mask & (plane0 | plane1)) {
						gfx_coll |= sbit;
						if (mdp & sbit) {
							plane0 &= ~fore_mask;	// Mask sprite if in background
							plane1 &= ~fore_mask;
						}
					}

					// Paint sprite
					plane0<<=xstart; plane1<<=xstart;
					for (i=xstart; i<xstop; i++, plane0<<=1, plane1<<=1) {
						if (plane1 & 0x80000000) {
							if (plane0 & 0x80000000)
								col = mm1_color;
							else
								col = color;
						} else {
							if (plane0 & 0x80000000)
								col = mm0_color;
							else
								continue;
						}
						PUT_PIXEL();
					}

				} else {			// Standard mode

					// Collision with graphics?
					if (fore_mask & sdata) {
						gfx_coll |= sbit;
						if (mdp & sbit)
							sdata &= ~fore_mask;	// Mask sprite if in background
					}
	
					// Paint sprite
					col = color;
					sdata<<=xstart;
					for (i=xstart; i<xstop; i++, sdata<<=1)
						if (sdata & 0x80000000) {
							PUT_PIXEL();
						}

				}
		}

	if (the_c64->ThePrefs.SpriteCollisions) {

		// Check sprite-sprite collisions
		if (clx_spr)
			clx_spr |= spr_coll;
		else {
			clx_spr |= spr_coll;
			irq_flag |= 0x04;
			if (irq_mask & 0x04) {
				irq_flag |= 0x80;
				the_cpu->TriggerVICIRQ();
			}
		}

		// Check sprite-background collisions
		if (clx_bgr)
			clx_bgr |= gfx_coll;
		else {
			clx_bgr |= gfx_coll;
			irq_flag |= 0x02;
			if (irq_mask & 0x02) {
				irq_flag |= 0x80;
				the_cpu->TriggerVICIRQ();
			}
		}
	}
}
